/*
This code is used for the dG calculations performed in the article titled 
"Impact of Primer Dimers and Self-Amplifying Hairpins on Reverse Transcription 
Loop-Mediated Isothermal Amplification Detection of Viral RNA"
Authors: Robert J. Meagher1 *, Aashish Priye1, Yooli K. Light2 Cheng Huang3, Eryu Wang4 
Code: Aashish Priye (apriye@sandia.gov) 
This program computes for a nucleotide sequence in a primer, the dG_init, the dG_terminal  
and the dG_hyb of the primers binding to its complementary template.  
*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include "export.h"
#include "calcul.h"

struct dG_analysis_term *get_results(struct sys_values *pst_sys_values){
    int i,j;			/* base pair counters */
    int i_sing_nuc;             /* sing_nuce detector */
    int i_freeend;              /* free end of nucleotide*/
    int i_length = 0;		/* length of the sequence */
    int i_proxoffset = 0;       /* current position from the sequence end*/
    int i_distoffset = 0;       /* current distance from the sequence end*/
    struct dG_analysis_term *pst_results; /* contains the results ... */

				/* initialisation of result variables */

    if ( (pst_results = (struct dG_analysis_term *)malloc(sizeof(struct dG_analysis_term))) == NULL){
      	 exit(EXIT_FAILURE);
    }
    pst_results->d_total_dG_init = 0.0;
    pst_results->d_total_dG_terminal = 0.0;
    for ( i = 0; i < NBNN ; i++)
	pst_results->i_crick[i] = 0;
    for ( i = 0; i < NBMM ; i++)
	pst_results->i_sing_nuc[i] = 0;
    for ( i = 0; i < NBDE ; i++)
	pst_results->i_freeends[i] = 0;
    pst_results->d_tm = 0.0;

    /*+------------------------------------------------------------------+
      | Length computation												 |
      +------------------------------------------------------------------+*/

    if (strlen(pst_sys_values->ps_sequence) > i_threshold)
	i_approx = TRUE;

    if (i_approx == TRUE)
	pst_results->d_tm = temp_est(pst_sys_values);

    /*+-----------------+
      | salt correction |
      +-----------------+*/

double temp_val(struct sys_values *pst_sys_values, struct dG_analysis_term *pst_results){

    double d_temp;		/* melting temperature */
    double salt_value = 0.0; /* ... */


    if (strncmp(pst_sys_values->s_sodium_correction,"wet91a",6) == 0)
	salt_value = 16.6 * log10 (pst_sys_values->d_conc_salt / (1.0 + 0.7 * pst_sys_values->d_conc_salt)) - 269.32;
    else 
	if (strncmp(pst_sys_values->s_sodium_correction,"san96a",6) == 0)
	    salt_value = 12.5 * log10 (pst_sys_values->d_conc_salt) - 273.15;
	else 
	    if (strncmp(pst_sys_values->s_sodium_correction,"san98a",6) == 0){
		salt_value = -273.15;
		pst_results->d_total_dG_terminal += 0.368 * (strlen(pst_sys_values->ps_sequence)-1) * log (pst_sys_values->d_conc_salt);
	    } else 
		if (strncmp(pst_sys_values->s_sodium_correction,"nak99a",6) == 0){
		    fprintf(ERROR," Sorry, not implemented yet\n");
		    exit(EXIT_FAILURE);
		}
				/* dG_analysis_term term */
    d_temp = pst_results->d_total_dG_init / (pst_results->d_total_dG_terminal + 1.987 * log (pst_sys_values->d_conc_probe/pst_sys_values->d_gnat))
				/* salt correction */
	+ salt_value;
    
    return d_temp;
}

import_from_file("primer_database.xlsx", "excel"); //Import primer sequences

struct nnset *read_nn(char *ps_nn_set, char *ps_path){
    struct nnset *pst_current_nn; /* pointer on a structure containing a set of nn_sys_values */
    FILE *pF_nn_file;		  /* handle of file containing a set of nn sys_values */
    char s_line[MAX_LINE];	  /* contains a line of a file or of stdin */
    char *pc_line_ptr;		  /* pointer moving along an input line */
    char *ps_nn_path;		  /* contains the address of the nn file */
    int i_crickcount = 0;	  /* counter of recorded crick's pairs */
    int i_count;
  
       /*+-----------------------------------------------------+
         | initialise a structure containing the nn sys_valueseters |
         +-----------------------------------------------------+*/

    pst_current_nn = (struct nnset *)calloc(1,sizeof(struct nnset));
    /* I AM SUPPOSED TO FREE THAT INSIDE THIS FUNCTION */

    if ((ps_nn_path = (char *)malloc(FILE_MAX)) == NULL){
      fprintf(ERROR," Function decode_input, line __LINE__:"
	      " Unable to allocate memory for the path of thenearest neighbor file.\n");
      exit(EXIT_FAILURE);
    }

    sprintf(ps_nn_path,"%s/%s",ps_path,ps_nn_set);
    /* construct the complete name of the nn set file */
    
    if ( (pF_nn_file = fopen(ps_nn_path,"r")) == NULL){
      /* cannot open file containing alternative nn set in this path */
      fprintf(ERROR," I was not able to open the file %s,\n"
	      " supposed to contain the set of nearest_neighbor sys_valueseters.\n",ps_nn_path);
      
      sprintf(ps_nn_path,"%s/%s",NN_BASE,ps_nn_set);
      
      if ( (pF_nn_file = fopen(ps_nn_path,"r")) == NULL){
	/* permits to check default directory if the env defined does not work */
	/* for instance -Hdnadna defined in default and -Afoo97a.nn define by env var*/
	fprintf(ERROR," I was not able to open the file %s,\n"
		" supposed to contain the set of nearest_neighbor sys_valueseters.\n",ps_nn_path);
	usage();
	exit(EXIT_FAILURE);
      }
      fprintf(ERROR," I am going to use the file %s instead.\n",ps_nn_path);
    }

       /*+-------------------------------------+
         | read the file containing the nn set |
         +-------------------------------------+*/
    i_count = 0;		/* initialise reference counter */
    while(!feof(pF_nn_file)){	/* read the file nn_file until it ends */
	fgets(s_line,sizeof(s_line),pF_nn_file);	/* read a line the file */
	pc_line_ptr = s_line;
	if(*pc_line_ptr == ' ')	/* skip uninformative spaces */
	    pc_line_ptr++;
	if(*pc_line_ptr == '/' || *pc_line_ptr == '\n')	/* skip empty line */
	    continue;
	if(*pc_line_ptr == 'R'){
	    if (i_count >= NUM_REF) /* In case there are too many references */
		continue;
/*	    pst_current_nn->s_reference[i_count] = (char *)malloc(MAX_REF); */
/* FIXME: has to allocate place for a new reference */
	    strncpy(pst_current_nn->s_reference[i_count],s_line,MAX_REF); /* enter the reference */
	    pst_current_nn->s_reference[i_count][MAX_REF - 1] = '\0'; /* security lock */
	    i_count++;		/* ready for next reference */
	}
	if(*pc_line_ptr == 'A'||*pc_line_ptr == 'a'
	   ||*pc_line_ptr == 'G'||*pc_line_ptr == 'g'
	   ||*pc_line_ptr == 'C'||*pc_line_ptr == 'c'
	   ||*pc_line_ptr == 'T'||*pc_line_ptr == 't'
	   ||*pc_line_ptr == 'U'||*pc_line_ptr == 'u'
	   ||*pc_line_ptr == 'I'||*pc_line_ptr == 'i'){
	    if (i_crickcount <= NBNN){
		sscanf(s_line,"%3s %lf %lf",pst_current_nn->ast_nndata[i_crickcount].s_crick_pair,
		       &(pst_current_nn->ast_nndata[i_crickcount].d_dG_init),
		       &(pst_current_nn->ast_nndata[i_crickcount].d_dG_terminal));
		i_crickcount++;
	    }else {
		fprintf(ERROR," I detected too many Crick's pairs in that file.\n"
			      " Only 16 pairs and two initiation factors are allowed.\n");
		exit(EXIT_FAILURE);
	    }
	}
    }
    fclose(pF_nn_file);
    return pst_current_nn;
}

    /*+-----------------------------------+
      | Computing nearest neighbor values |
      +-----------------------------------+*/

    else {
	
      if ( *(pst_sys_values->ps_sequence) == '-' || *(pst_sys_values->ps_complement) == '-'){
	i_freeend = TRUE;
	if (i_dnadna == FALSE && i_alt_de == FALSE){
	 
	    }
	    i_proxoffset++;
	    for (i = 0 ; i < NBDE ; i++){ /* find the free-end term */
	      if ( (strncmp(pst_sys_values->ps_sequence,pst_sys_values->pst_present_de->ast_dedata[i].s_crick_pair,2) == 0)
		   && (strncmp(pst_sys_values->ps_complement,&pst_sys_values->pst_present_de->ast_dedata[i].s_crick_pair[3],2) == 0)){
		pst_results->d_total_dG_init += pst_sys_values->pst_present_de->ast_dedata[i].d_dG_init;
		pst_results->d_total_dG_terminal += pst_sys_values->pst_present_de->ast_dedata[i].d_dG_terminal;
		pst_results->i_freeends[i]++;
		i_freeend = FALSE;
	      }
	    }
	    if (i_freeend == TRUE){	      
	      exit(EXIT_FAILURE);
	    }
	}	
	
	if ( *(pst_sys_values->ps_sequence + strlen(pst_sys_values->ps_sequence)-1) == '-' || *(pst_sys_values->ps_complement + strlen(pst_sys_values->ps_complement)-1) == '-'){
	  i_freeend = TRUE;
	  i_distoffset++;
	  for (i = 0 ; i < NBDE ; i++){ /* find the  free end term */
	    if ( (strncmp(pst_sys_values->ps_sequence+strlen(pst_sys_values->ps_sequence)-2,pst_sys_values->pst_present_de->ast_dedata[i].s_crick_pair,2) == 0)
		 && (strncmp(pst_sys_values->ps_complement+strlen(pst_sys_values->ps_sequence)-2,&pst_sys_values->pst_present_de->ast_dedata[i].s_crick_pair[3],2) == 0)){
	      pst_results->d_total_dG_init += pst_sys_values->pst_present_de->ast_dedata[i].d_dG_init;
	      pst_results->d_total_dG_terminal += pst_sys_values->pst_present_de->ast_dedata[i].d_dG_terminal;
	      pst_results->i_freeends[i]++;
	      i_freeend = FALSE;
	    }
	  }
	  if (i_freeend == TRUE){
	    fprintf(ERROR,"nearest neighbor sys_valueseters for %c%c/%c%c not found. Check the file containing\n"
		    " the information on dangling ends\n",*(pst_sys_values->ps_sequence+strlen(pst_sys_values->ps_sequence)-2),*(pst_sys_values->ps_sequence+strlen(pst_sys_values->ps_sequence)-1),*(pst_sys_values->ps_complement+strlen(pst_sys_values->ps_sequence)-2),*(pst_sys_values->ps_complement+strlen(pst_sys_values->ps_sequence)-1));
	    exit(EXIT_FAILURE);
	  }
	}
				/* determination of initiation terms*/
	if ( *(pst_sys_values->ps_sequence + i_proxoffset) == 'A' || *(pst_sys_values->ps_sequence + i_proxoffset) == 'T')
	    for(i = 0 ; i < NBNN ; i++) /* find the initiation term */
		if ( strncmp(pst_sys_values->pst_present_nn->ast_nndata[i].s_crick_pair,"IA",2) == 0){
		    pst_results->d_total_dG_init += pst_sys_values->pst_present_nn->ast_nndata[i].d_dG_init;
		    pst_results->d_total_dG_terminal += pst_sys_values->pst_present_nn->ast_nndata[i].d_dG_terminal;
		}
	if ( *(pst_sys_values->ps_sequence + i_proxoffset) == 'G' || *(pst_sys_values->ps_sequence + i_proxoffset) == 'C')
	    for(i = 0 ; i < NBNN ; i++) /* find the initiation term */
		if ( strncmp(pst_sys_values->pst_present_nn->ast_nndata[i].s_crick_pair,"IG",2) == 0){
		    pst_results->d_total_dG_init += pst_sys_values->pst_present_nn->ast_nndata[i].d_dG_init;
		    pst_results->d_total_dG_terminal += pst_sys_values->pst_present_nn->ast_nndata[i].d_dG_terminal;
		}
				/* determination of initiation terms*/
	if ( *(pst_sys_values->ps_sequence + strlen(pst_sys_values->ps_sequence)-1-i_distoffset) == 'A' || *(pst_sys_values->ps_sequence + strlen(pst_sys_values->ps_sequence)-1-i_distoffset) == 'T')
	    for(i = 0 ; i < NBNN ; i++) /* find the initiation term */
		if ( strncmp(pst_sys_values->pst_present_nn->ast_nndata[i].s_crick_pair,"IA",2) == 0){
		    pst_results->d_total_dG_init += pst_sys_values->pst_present_nn->ast_nndata[i].d_dG_init;
		    pst_results->d_total_dG_terminal += pst_sys_values->pst_present_nn->ast_nndata[i].d_dG_terminal;
		}
	if ( *(pst_sys_values->ps_sequence + strlen(pst_sys_values->ps_sequence)-1-i_distoffset) == 'G' || *(pst_sys_values->ps_sequence + strlen(pst_sys_values->ps_sequence)-1-i_distoffset) == 'C')
	    for(i = 0 ; i < NBNN ; i++) /* find the initiation term */
		if ( strncmp(pst_sys_values->pst_present_nn->ast_nndata[i].s_crick_pair,"IG",2) == 0){
		    pst_results->d_total_dG_init += pst_sys_values->pst_present_nn->ast_nndata[i].d_dG_init;
		    pst_results->d_total_dG_terminal += pst_sys_values->pst_present_nn->ast_nndata[i].d_dG_terminal;
		}
				/* Travel through the sequence */
	if (strlen(pst_sys_values->ps_sequence)-1 <= 0){
	    fprintf(ERROR," Oups, the lengh of the sequence seems zero or less ...\n");
	    exit(1);
	}

	i_length = strlen(pst_sys_values->ps_sequence)-1 -i_proxoffset -i_distoffset;
	for (i = 0 + i_proxoffset ; i < i_length; i++){
	    i_sing_nuc = FALSE;
	    switch (pst_sys_values->ps_sequence[i]) {
		case 'A':
		    if (pst_sys_values->ps_complement[i] != 'T')
			i_sing_nuc = TRUE;
		    break;
		case 'G':
		    if (pst_sys_values->ps_complement[i] != 'C')
			i_sing_nuc = TRUE;
		    break;
		case 'C':
		    if (pst_sys_values->ps_complement[i] != 'G')
			i_sing_nuc = TRUE;
		    break;
		case 'T':
		    if (pst_sys_values->ps_complement[i] != 'A')
			i_sing_nuc = TRUE;
		    break;
		default:
		  
		    exit(EXIT_FAILURE);
	    }
	    switch (pst_sys_values->ps_sequence[i+1]) {
		case 'A':
		    if (pst_sys_values->ps_complement[i+1] != 'T')
			i_sing_nuc = TRUE;
		    break;
		case 'G':
		    if (pst_sys_values->ps_complement[i+1] != 'C')
			i_sing_nuc = TRUE;
		    break;
		case 'C':
		    if (pst_sys_values->ps_complement[i+1] != 'G')
			i_sing_nuc = TRUE;
		    break;
		case 'T':
		    if (pst_sys_values->ps_complement[i+1] != 'A'){
			i_sing_nuc = TRUE;
		    }
		    break;
		default:
		   
	    }
	    if (i_sing_nuc == TRUE){
		if (i == (0 + i_proxoffset) || i == (i_length - 1)){
		  
		    exit(EXIT_FAILURE);
		}
		if (i_dnadna == FALSE && i_alt_mm == FALSE){
	
		}
				/*comparing bases one by one*/
		for (j = 0 ; j < NBMM ; j++){
		  if ( (strncmp(&pst_sys_values->ps_sequence[i],pst_sys_values->pst_present_mm->ast_mmdata[j].s_crick_pair,2) == 0)
		       && (strncmp(&pst_sys_values->ps_complement[i],&pst_sys_values->pst_present_mm->ast_mmdata[j].s_crick_pair[3],2) == 0)){
		    pst_results->i_sing_nuc[j]++;
		    if (pst_sys_values->pst_present_mm->ast_mmdata[j].d_dG_init != 99999){
		      pst_results->d_total_dG_init += pst_sys_values->pst_present_mm->ast_mmdata[j].d_dG_init;
		      pst_results->d_total_dG_terminal += pst_sys_values->pst_present_mm->ast_mmdata[j].d_dG_terminal;
		      i_sing_nuc = FALSE; /*nearest neighbor for sing_nuc identified */
		    }
		    break;
		  }
		}
		if (i_sing_nuc == TRUE){
		  fprintf(ERROR,"nearest neighbor sys_valueseters for %c%c/%c%c not found. Check the file containing\n"
			  " the information on sing_nuces\n",pst_sys_values->ps_sequence[i],pst_sys_values->ps_sequence[i+1],pst_sys_values->ps_complement[i],pst_sys_values->ps_complement[i+1]);
		  exit(EXIT_FAILURE);
		}
	    } else 
		/*regular pair comparison*/
		for ( j = 0; j < NBNN; j++)
		    if ( strncmp(&pst_sys_values->ps_sequence[i],pst_sys_values->pst_present_nn->ast_nndata[j].s_crick_pair,2) == 0){
			pst_results->i_crick[j]++;
			pst_results->d_total_dG_init += pst_sys_values->pst_present_nn->ast_nndata[j].d_dG_init;
			pst_results->d_total_dG_terminal += pst_sys_values->pst_present_nn->ast_nndata[j].d_dG_terminal;
		    }
	}
	pst_results->d_tm = temp_val(pst_sys_values,pst_results);
    }
    return pst_results;
}


double temp_est(struct sys_values *pst_sys_values){
    double d_temp;		/* melting temperature */
    int i_size;			/* size of the duplex */
    int i_numbergc;		
    double  gc_content;		
    char * seq_search;     	/* translate the sequence */

  

    i_size = strlen(pst_sys_values->ps_sequence);
    if (i_size == 0){
	fprintf(ERROR," The size of the duplex appears to be null. Therefore I\n"
                      " cannot compute approximation of the melting temperature.\n");
	exit(EXIT_FAILURE);
    }

    /*+----------------+
      | percent of G+C |
      +----------------+*/
	
     seq_search = pst_sys_values->ps_sequence;
    i_numbergc = 0;
    while(* seq_search != '\0'){
	if (* seq_search == 'G' || * seq_search == 'C')
	    i_numbergc++;
	 seq_search++;
    }
     gc_content = ( (double)i_numbergc / (double)i_size ) * 100;

    /*+---------------------+
      | melting temperature |
      +---------------------+*/

    if(i_dnadna == TRUE)
	d_temp = 81.5 
	    + 16.6 * log10(pst_sys_values->d_conc_salt / (1.0 + 0.7 * pst_sys_values->d_conc_salt)) 
	    + 0.41 *  gc_content
	    - 500.0 / (double)i_size;
    else 
	if(i_dnarna == TRUE)
	    d_temp = 67 
		+ 16.6 * log10(pst_sys_values->d_conc_salt / (1.0 + 0.7 * pst_sys_values->d_conc_salt)) 
		+ 0.8 *  gc_content
		- 500.0 / (double)i_size;
	else 
	    if(i_rnarna == TRUE)
		d_temp = 78 
		    + 16.6 * log10(pst_sys_values->d_conc_salt / (1.0 + 0.7 * pst_sys_values->d_conc_salt)) 
		    + 0.8 *  gc_content
		    - 500.0 / (double)i_size;
	    else{
		fprintf(ERROR," I do not find any hybridisation type and therefore\n"
                              " I cannot compute the approximative melting temperature\n");
		exit(EXIT_FAILURE);
	    }
    
    return d_temp;
}


/******************************************
 * Construct the complement of a sequence *
 ******************************************/

char *make_complement(char *ps_sequence){
  char *ps_complement;	       /* computed complement of the sequence input */
  char *pc_base_seq;           /* moving pointer on sequence */
  char *pc_base_comp;	       /* moving pointer on complement */
  
  if ( (ps_complement = (char *)malloc(strlen(ps_sequence)+1)) == NULL){
    fprintf(ERROR," function make_complement, line __LINE__:\n"
	    "Unable to allocate memory to register the sequence\n");
    exit(EXIT_FAILURE);
  }
  pc_base_seq = ps_sequence;	/* pointer on the beginning of sequence */
  pc_base_comp = ps_complement;
  while (*pc_base_seq != '\0'){
    switch (*pc_base_seq){
    case 'A':
      *pc_base_comp = 'T';
      break;
    case 'G':
      *pc_base_comp = 'C';
      break;
    case 'C':
      *pc_base_comp = 'G';
      break;
    case 'T':
      *pc_base_comp = 'A';
      break;
    case '-':
      *pc_base_comp = '-';
    default:
      fprintf(ERROR," It seems that one base of sequence is illegal.\n"
	      " I cannot compute the complement.\n");
    }
    pc_base_seq++;
    pc_base_comp++;
  }
  *pc_base_comp = '\0';
  return ps_complement;
}

/************************************
 * Check the legality of a sequence *
 ************************************/

int check_sequence(char *ps_sequence){
  char *pc_base;                /*moving pointer on base*/
  int i_mistakes = 0;           /*error counter*/

  pc_base = ps_sequence;

  while (*pc_base != '\0'){
    /* capitalise the bases */
      *pc_base = toupper((int)*pc_base);

				/* change uridine into thymidine */
      if (*pc_base == 'U')
	  *pc_base = 'T';
      if (*pc_base != 'A' && *pc_base != 'G' && *pc_base != 'C' && *pc_base != 'T' && *pc_base != '-')
	  i_mistakes++;
      pc_base++;
  }
  return(i_mistakes);
}


export_to_file(get_results, pst_sys_values, "csv"); //Export dG values to excel file
export_to_file(pst_sys_values, "csv"); //Export dG values to excel file


